#include "attacker.h"

#define ATTACKER_SPRITE_SIZE 48

#define FIRST_MOVEMENT_FRAME 0
#define LAST_MOVEMENT_FRAME 1
#define FIRST_EXPLODING_FRAME 2
#define LAST_EXPLODING_FRAME 4

#define VIDEO_FRAMES_PER_ANIMATION_FRAME 3

#define VIDEO_FRAMES_BETWEEN_MOVEMENTS 1
#define PIXELS_PER_MOVEMENT 2
#define MOVEMENT_SCREEN_BOUNDARY_HORIZONTAL_PIXELS 10
#define MOVEMENT_SCREEN_BOUNDARY_VERTICAL_PIXELS 10

#define MIN_FRAMES_BEFORE_NEXT_DECISION 15
#define MAX_FRAMES_BEFORE_NEXT_DECISION 90

#define ATTACKER_MAX_LIFE 8

// out of 100
#define CHANCE_TO_MOVE_RIGHT 30
#define CHANCE_TO_MOVE_LEFT 30

static char* p_attackerFilepath = "attacker.anim";

Animatable p_attacker;
int32 p_attackerMovementCounterVideoFrames;
bool p_attackerMovementIsToTheRight;
bool p_isMoving;
int32 p_attackerFramesLeftBeforeDecision;
int p_bulletHandle;
bool (*p_attackerCheckCollisionCallback)( Rectangle position );

int p_life;

void p_rollNewDecisionFrameAmount() {
	p_attackerFramesLeftBeforeDecision = MIN_FRAMES_BEFORE_NEXT_DECISION + 
		( rand() % ( MAX_FRAMES_BEFORE_NEXT_DECISION - MIN_FRAMES_BEFORE_NEXT_DECISION ) );
}

void p_attackerDecision() {
	int decision;
	
	// keep doing what already doing
	if( p_attackerFramesLeftBeforeDecision > 0 ) {
		p_attackerFramesLeftBeforeDecision--;
		return;
	}
	
	p_rollNewDecisionFrameAmount();
	decision = rand() % 100;
	
	if( decision < CHANCE_TO_MOVE_RIGHT ) {
		p_attackerMovementIsToTheRight = TRUE;
		p_isMoving = TRUE;
		return;
	}
	decision -= CHANCE_TO_MOVE_RIGHT;
	
	if( decision < CHANCE_TO_MOVE_LEFT ) {
		p_attackerMovementIsToTheRight = FALSE;
		p_isMoving = TRUE;
		return;
	}
	
	// stand still, but not for too long
	p_attackerFramesLeftBeforeDecision = p_attackerFramesLeftBeforeDecision / 2;
	p_isMoving = FALSE;	
}

void p_checkCollisionAndHideBullet( Rectangle position ) {
	if( p_attackerCheckCollisionCallback( position ) ) {
		bullets_cancel( p_bulletHandle );
	}
}

void p_fire() {
	bullets_fire( 
		p_bulletHandle, 
		FAST,
		UP, 
		p_attacker.x + ATTACKER_SPRITE_SIZE / 2 - BULLET_SPRITE_SIZE / 2,
		p_attacker.y,
		p_checkCollisionAndHideBullet
		);
}

void p_moveAttacker() {
	int movementDelta = PIXELS_PER_MOVEMENT;
	bool isPastLeftBoundary = 
		p_attacker.x < MOVEMENT_SCREEN_BOUNDARY_HORIZONTAL_PIXELS;
	bool isPastRightBoundary = 
		p_attacker.x + ATTACKER_SPRITE_SIZE > 
		display_getScreenWidth() - MOVEMENT_SCREEN_BOUNDARY_HORIZONTAL_PIXELS;
		
	if( p_attacker.state != ALIVE ) {
		return;
	}
	
	if( !p_isMoving ) {
		p_fire();
		return;
	}

	p_attackerMovementCounterVideoFrames++;
	if( p_attackerMovementCounterVideoFrames >= VIDEO_FRAMES_BETWEEN_MOVEMENTS ) {
		p_attackerMovementCounterVideoFrames = 0;
	}
	
	if( !p_attackerMovementIsToTheRight && isPastLeftBoundary ) {
		p_attackerMovementIsToTheRight = TRUE;
	}
	if( p_attackerMovementIsToTheRight && isPastRightBoundary ) {
		p_attackerMovementIsToTheRight = FALSE;
	}
	
	if( !p_attackerMovementIsToTheRight ) {
		movementDelta = -movementDelta;
	}
	
	if( p_attackerMovementCounterVideoFrames == 0 ) {
		p_attacker.x += movementDelta;
	}

	sprite_moveTo( p_attacker.sprite, p_attacker.x, p_attacker.y );
	
	p_fire();
}

void p_animateAttacker() {
	if( p_attacker.state == DEAD ) {
		return;
	}
	
	animatables_animate( &p_attacker, VIDEO_FRAMES_PER_ANIMATION_FRAME );
}

void attacker_loadDefaults() {
	p_isMoving = TRUE;
	p_attackerMovementCounterVideoFrames = 0;
	p_attackerMovementIsToTheRight = TRUE;
	p_life = ATTACKER_MAX_LIFE;
	p_attacker.state = ALIVE;
	p_attacker.firstFrame = FIRST_MOVEMENT_FRAME;
	p_attacker.lastFrame = LAST_MOVEMENT_FRAME;
	p_attacker.currentFrame = FIRST_MOVEMENT_FRAME;
	p_attacker.videoFrame = 0;
	
	sprite_moveTo( p_attacker.sprite, p_attacker.x, p_attacker.y );
	sprite_setFrame( p_attacker.sprite, p_attacker.currentFrame );
}

void attacker_init(
	bool (*checkCollisionCallback)( Rectangle position ) 
	) {
		
	p_attackerCheckCollisionCallback = checkCollisionCallback;
	p_attacker.sprite = sprite_load( p_attackerFilepath );
	p_attacker.spriteIndex = sprite_allocateNextIndex();
	sprite_setFrame( p_attacker.sprite, 0 );
	
	p_attacker.y = display_getScreenHeight() - ATTACKER_SPRITE_SIZE - MOVEMENT_SCREEN_BOUNDARY_VERTICAL_PIXELS;	
	p_attacker.x = display_getScreenWidth() / 2 - ATTACKER_SPRITE_SIZE / 2;	

	display_displayListAdd( 
		(Node*)p_attacker.sprite, 
		TRUE, 
		p_attacker.spriteIndex
		);
		
	p_bulletHandle = bullets_allocate();
	
	attacker_loadDefaults();
}

void attacker_refresh() {
	p_attackerDecision();
	p_moveAttacker();
	p_animateAttacker();
	
	if( p_life == 0 ) {
		sprite_setFrame( p_attacker.sprite, FIRST_EXPLODING_FRAME );
	}
}

bool attacker_checkCollision( Rectangle position ) {
	Rectangle hitbox = p_attacker.sprite->position;
	hitbox.topLeft.pt_X += 3;
	hitbox.bottomRight.pt_X -= 3;
	if( sprite_isRectangleOverlap( &position, &hitbox ) ) {
		p_life--;
		return TRUE;
	}

	return FALSE;
}

int attacker_getLife() {
	return p_life;
}
