#include "bullets.h"

#define NUM_BULLETS 2

#define FAST_SPEED 5
#define SLOW_SPEED 3

#define BULLET_SCREEN_BOUNDARY 5

#define VIDEO_FRAMES_PER_ANIMATION_FRAME 2

#define FIRST_MOVEMENT_FRAME 0
#define LAST_MOVEMENT_FRAME 1

typedef struct Bullet {
	Animatable animatable;
	Direction direction;
	int pixelsPerFrame;
	void (*positionChangedCallback)( Rectangle position );
} Bullet;

static char* p_bulletFilepath = "bullet.anim";
Bullet p_bullets[NUM_BULLETS];
int p_nextAllocatedBullet = 0;

void p_moveBullet( Bullet* bullet, int x, int y ) {
	Rectangle hitbox;
	
	bullet->animatable.x = x;
	bullet->animatable.y = y;
	sprite_moveTo( bullet->animatable.sprite, bullet->animatable.x, bullet->animatable.y );
	
	if( bullet->animatable.state != ALIVE ) {
		return;
	}
	
	// slimmer hitbox than the entire bullet sprite
	hitbox = bullet->animatable.sprite->position;
	hitbox.topLeft.pt_X += BULLET_SPRITE_SIZE / 3;
	hitbox.bottomRight.pt_X -= BULLET_SPRITE_SIZE / 3;
	
	bullet->positionChangedCallback( hitbox );
}

void p_hideBullet( Bullet* bullet ) {
	bullet->animatable.state = DEAD;
	p_moveBullet( bullet, display_getScreenWidth(), 0 );
	sprite_moveTo( bullet->animatable.sprite, bullet->animatable.x, bullet->animatable.y );
}

void p_moveBullets() {
	int i;
	Bullet* current;
	
	for( i = 0; i < NUM_BULLETS; i++ ) {
		current = &p_bullets[i];
			
		if( current->animatable.state != ALIVE ) {
			continue;
		}
		
		switch( current->direction ) {
			case UP:
				if( current->animatable.y > BULLET_SCREEN_BOUNDARY ) {
					p_moveBullet( current, current->animatable.x, current->animatable.y - current->pixelsPerFrame );
					continue;
				}
				p_hideBullet( current );
				break;
			case DOWN:
				if( current->animatable.y < display_getScreenHeight() - BULLET_SCREEN_BOUNDARY - BULLET_SPRITE_SIZE ) {
					p_moveBullet( current, current->animatable.x, current->animatable.y + current->pixelsPerFrame );
					continue;
				}
				p_hideBullet( current );
				break;
		}
	}
}

void p_animateBullets() {
	int i;
	Bullet* current;
	
	for( i = 0; i < NUM_BULLETS; i++ ) {
		current = &p_bullets[i];
			
		if( current->animatable.state != ALIVE ) {
			continue;
		}
		
		animatables_animate( &current->animatable, VIDEO_FRAMES_PER_ANIMATION_FRAME );
	}
}

void bullets_refresh() {
	p_moveBullets();
	p_animateBullets();
}

void bullets_cancel( int which ) {
	p_hideBullet( &p_bullets[which] );
}

void bullets_loadDefaults() {
	int i;
	Animatable* current;
	
	for( i = 0; i < NUM_BULLETS; i++ ) {
		current = &p_bullets[i].animatable;
		
		current->state = DEAD;
		p_hideBullet( &p_bullets[i] );
		
		current->firstFrame = FIRST_MOVEMENT_FRAME;
		current->lastFrame = LAST_MOVEMENT_FRAME;
		current->currentFrame = FIRST_MOVEMENT_FRAME;
		current->videoFrame = 0;
	}
}

void bullets_fire( 
	int which, 
	Speed speed, 
	Direction direction, 
	int startX, 
	int startY, 
	void (*positionChangedCallback)( Rectangle position ) 
	) {
	// can only fire an inactive bullet
	if( p_bullets[which].animatable.state != DEAD ) {
		return;
	}
	
	p_bullets[which].pixelsPerFrame = speed == FAST ? FAST_SPEED : SLOW_SPEED;
	p_bullets[which].direction = direction;
	p_bullets[which].animatable.state = ALIVE;
	p_bullets[which].positionChangedCallback = positionChangedCallback;
	p_moveBullet( &p_bullets[which], startX, startY );
}

int bullets_allocate() {
	return p_nextAllocatedBullet++;
}

void bullets_init() {
	int i;
	Animatable* current;
	
	for( i = 0; i < NUM_BULLETS; i++ ) {
		current = &p_bullets[i].animatable;
		
		current->sprite = sprite_load( p_bulletFilepath );
		current->spriteIndex = sprite_allocateNextIndex();
		sprite_setFrame( current->sprite, 0 );
		
		display_displayListAdd( 
			(Node*)current->sprite, 
			TRUE, 
			current->spriteIndex
			);
	}
	
	bullets_loadDefaults();
}