#include "display.h"

#define	MAX_SCREENS 2
#define	GRAPHICS_BASE_PATH_PAL1 "PAL/Graphics/" // 320x288 pixels - 50 frames/sec
#define	GRAPHICS_BASE_PATH_PAL2 "PAL/Graphics/" // 384x288 pixels - 50 frames/sec.
#define	GRAPHICS_BASE_PATH_NTSC "NTSC/Graphics/" // 320x240 pixels - 60 frames/sec.

typedef enum DisplayType {
	PAL1 = 1,
	PAL2 = 2,
	NTSC = 0
} DisplayType;

DisplayType p_displayType;
Item p_VRAMIORequest = -1;
ScreenContext* p_screenContext = NULL;
int32 p_currentScreenIndex = 0;
ubyte* p_backgroundImage = NULL;
List* p_spritesList = NULL;

void display_init() {
	int32 displayType;
	
	OpenGraphicsFolio();
	p_VRAMIORequest = CreateVRAMIOReq();

	// allocate memory for the screen context
	p_screenContext = (ScreenContext *)AllocMem( sizeof( ScreenContext ), MEMTYPE_ANY );

	displayType = GetDisplayType();
	
	if( displayType == DI_TYPE_PAL1 ) {
		p_displayType = PAL1;
	} else if( displayType == DI_TYPE_PAL2 ) {
		p_displayType = PAL2;
	} else {
		p_displayType = NTSC;
	}	
	
	// create enough screens to be able to flip them (double buffering)
	CreateBasicDisplay( p_screenContext, displayType, MAX_SCREENS );

	// allocate memory for the list of sprites, and initialize
	p_spritesList = (List*)AllocMem( sizeof( List ), MEMTYPE_ANY );
	InitList( p_spritesList, "p_spritesList" );
}

void p_swapBuffers() {
	DisplayScreen( p_screenContext->sc_Screens[p_currentScreenIndex], 0 );
	
	// next screen
	p_currentScreenIndex++;
	if( p_currentScreenIndex >= MAX_SCREENS ) {
		p_currentScreenIndex=0;
	}
}

void display_render() {
	CCB* ccb;
	Point rectangleCorners[4];
	Sprite* sprite;
	
	// background image
	CopyVRAMPages( 
		p_VRAMIORequest, 
		p_screenContext->sc_Bitmaps[p_currentScreenIndex]->bm_Buffer, 
		p_backgroundImage, 
		p_screenContext->sc_nFrameBufferPages, 
		0xFFFFFFFF 
		);
	
	if( !IsListEmpty( p_spritesList ) ){
		
		// for each sprite in the sprites list, render it
		ScanList( p_spritesList, sprite, Sprite ) {
		
			sprite->data->cur_Frame=Convert32_F16( sprite->animationFrame );
			ccb = GetAnimCel( sprite->data, 0 );
			
			// four corners
			rectangleCorners[0].pt_X = sprite->position.topLeft.pt_X;
			rectangleCorners[0].pt_Y = sprite->position.topLeft.pt_Y;
			rectangleCorners[1].pt_X = sprite->position.bottomRight.pt_X;
			rectangleCorners[1].pt_Y = sprite->position.topLeft.pt_Y;
			rectangleCorners[2].pt_X = sprite->position.bottomRight.pt_X;
			rectangleCorners[2].pt_Y = sprite->position.bottomRight.pt_Y;
			rectangleCorners[3].pt_X = sprite->position.topLeft.pt_X;
			rectangleCorners[3].pt_Y = sprite->position.bottomRight.pt_Y;
			
			MapCel( ccb, rectangleCorners );
			LAST_CEL( ccb );
			DrawCels( p_screenContext->sc_BitmapItems[p_currentScreenIndex], ccb );
		}
	}

	p_swapBuffers();
}

void display_loadBackground( char *filename ) {
	char basePath[MAX_PATH];
	
	if( p_displayType == PAL1 ) {
		strcpy( basePath, GRAPHICS_BASE_PATH_PAL1 );
	} else if( p_displayType == PAL2 ) {
		strcpy( basePath, GRAPHICS_BASE_PATH_PAL2 );
	} else{
		strcpy( basePath, GRAPHICS_BASE_PATH_NTSC );
	}
	
	strcat( basePath, filename );
	
	if( p_backgroundImage != NULL ) {
		FreeMem( p_backgroundImage, p_screenContext->sc_nFrameByteCount );
		p_backgroundImage = NULL;
	}
	
	// allocate memory for the background image
	p_backgroundImage = (ubyte *)AllocMem(
		(int)( p_screenContext->sc_nFrameByteCount ), 
		MEMTYPE_STARTPAGE | MEMTYPE_VRAM | MEMTYPE_CEL
		);
	
	// load into VRAM
	SetVRAMPages( p_VRAMIORequest, p_backgroundImage, 0, p_screenContext->sc_nFrameBufferPages, -1);
	LoadImage(basePath, p_backgroundImage, (VdlChunk **)NULL, p_screenContext);
}


void display_displayListAdd( Node* node, bool appendAfterTail, int32 index ){
	if( appendAfterTail ) {
		InsertNodeFromTail( p_spritesList, node );
	} else {
		InsertNodeFromHead( p_spritesList, node );
	}
	
	// raise priority of new node
	SetNodePri( node, index );
}

void display_displayListRemove( Node* node ) {
	RemNode( node );
}

void display_emptyDisplayList() {
	Node* node = NULL;

	if( IsListEmpty( p_spritesList ) ) {
		return;
	}
	
	ScanList( p_spritesList, node, Node ) {
		RemNode( node );
	}
}

int32 display_displayListCount() {
	return GetNodeCount( p_spritesList );
}

int display_getScreenWidth() {
	switch( p_displayType ) {
		case PAL1:
			return 320;
		case PAL2:
			return 384;
		case NTSC:
			return 320;
	}
	
	return 0;
}

int display_getScreenHeight() {
	switch( p_displayType ) {
		case PAL1:
			return 288;
		case PAL2:
			return 288;
		case NTSC:
			return 240;
	}
	
	return 0;
}