#include "invaders.h"

#define NUM_INVADERS_VERTICAL 4
#define NUM_INVADERS_HORIZONTAL 8
#define INVADER_SPRITE_SIZE 20
#define BUFFER_BETWEEN_INVADER_SPRITES 5

#define FIRST_MOVEMENT_FRAME 0
#define LAST_MOVEMENT_FRAME 1
#define FIRST_EXPLODING_FRAME 2
#define LAST_EXPLODING_FRAME 4

#define NUM_INVADER_TEXTURES 2

#define VIDEO_FRAMES_PER_ANIMATION_FRAME 7
#define EXPLODING_VIDEO_FRAMES_PER_ANIMATION_FRAME 15

#define VIDEO_FRAMES_BETWEEN_MOVEMENTS 3
#define PIXELS_PER_MOVEMENT 1
#define MOVEMENT_SCREEN_BOUNDARY_HORIZONTAL_PIXELS 10
#define MOVEMENT_SCREEN_BOUNDARY_VERTICAL_PIXELS 10

#define CURSOR_FRAME_COUNT_TO_STAY_PUT 80
#define CURSOR_SPRITE_SIZE 28

static char* p_invaderFilepaths[NUM_INVADER_TEXTURES] = { "invader.anim", "invader2.anim" };
static char* p_cursorFilepath = "cursor.cel";

Animatable p_invaders[NUM_INVADERS_VERTICAL][NUM_INVADERS_HORIZONTAL];
int32 p_movementCounterVideoFrames;
bool p_movementIsToTheRight;

int p_invaderBulletHandle;
bool (*p_invadersCheckCollisionCallback)( Rectangle position );

Sprite* p_infoPanel;
static char* p_infoPanelFilepath = "info.cel";

Sprite* p_cursor;
int32 p_cursorRow, p_cursorCol;
int32 p_cursorFramesSinceMoved;

void p_moveInvaders() {
	int row, col;
	Animatable* current;
	int movementDelta = PIXELS_PER_MOVEMENT;
	int leftMostInvaderEdge = display_getScreenWidth();
	int rightMostInvaderEdge = 0;
	bool isPastLeftBoundary;
	bool isPastRightBoundary;

	for( row = 0; row < NUM_INVADERS_VERTICAL; row++ ) {
		for( col = 0; col < NUM_INVADERS_HORIZONTAL; col++ ) {
			if( p_invaders[row][col].state != ALIVE ) {
				continue;
			}
			
			if( p_invaders[row][col].x < leftMostInvaderEdge ) {
				leftMostInvaderEdge = p_invaders[row][col].x;
			}
			if( p_invaders[row][col].x + INVADER_SPRITE_SIZE > rightMostInvaderEdge ) {
				rightMostInvaderEdge = p_invaders[row][col].x + INVADER_SPRITE_SIZE;
			}
		}
	}
	isPastLeftBoundary = leftMostInvaderEdge < MOVEMENT_SCREEN_BOUNDARY_HORIZONTAL_PIXELS;
	isPastRightBoundary = rightMostInvaderEdge > display_getScreenWidth() - MOVEMENT_SCREEN_BOUNDARY_HORIZONTAL_PIXELS;
		
	p_movementCounterVideoFrames++;
	if( p_movementCounterVideoFrames >= VIDEO_FRAMES_BETWEEN_MOVEMENTS ) {
		p_movementCounterVideoFrames = 0;
	}
	
	if( !p_movementIsToTheRight && isPastLeftBoundary ) {
		p_movementIsToTheRight = TRUE;
	}
	if( p_movementIsToTheRight && isPastRightBoundary ) {
		p_movementIsToTheRight = FALSE;
	}
	
	if( !p_movementIsToTheRight ) {
		movementDelta = -movementDelta;
	}
	
	for( row = 0; row < NUM_INVADERS_VERTICAL; row++ ) {
		for( col = 0; col < NUM_INVADERS_HORIZONTAL; col++ ) {
			current = &p_invaders[row][col];
			
			if( current->state != ALIVE ) {
				continue;
			}
			
			if( p_movementCounterVideoFrames == 0 ) {
				current->x += movementDelta;
			}
		
			sprite_moveTo( current->sprite, current->x, current->y );
		}
	}
}

void p_explodeInvader( Animatable* invader ) {
	invader->state = EXPLODING;
	invader->firstFrame = FIRST_EXPLODING_FRAME;
	invader->lastFrame = LAST_EXPLODING_FRAME;
	invader->currentFrame = FIRST_EXPLODING_FRAME;
	sprite_setFrame( invader->sprite, invader->currentFrame );
	// start at 1 so we can check when we're back to 
	// animation frame 0, video frame 0 to stop explosion and erase the invader
	invader->videoFrame = 1;
}

void p_animateInvaders() {
	int row, col;
	Animatable* current;
	int videoFramesPerAnimationFrame = VIDEO_FRAMES_PER_ANIMATION_FRAME;
		
	for( row = 0; row < NUM_INVADERS_VERTICAL; row++ ) {
		for( col = 0; col < NUM_INVADERS_HORIZONTAL; col++ ) {
		
			current = &p_invaders[row][col];
			
			if( current->state == DEAD ) {
				continue;
			}
			
			if( current->state == EXPLODING ) {
				videoFramesPerAnimationFrame = EXPLODING_VIDEO_FRAMES_PER_ANIMATION_FRAME;
			}
			
			animatables_animate( current, videoFramesPerAnimationFrame );
			
			if( current->state == EXPLODING && 
				current->currentFrame == current->firstFrame &&
				current->videoFrame == 0
				) {
				animatables_kill( current );
			}
		}
	}	
}

int32 invaders_countRemaining() {
	int row, col;
	int result = 0;
	
	for( row = 0; row < NUM_INVADERS_VERTICAL; row++ ) {
		for( col = 0; col < NUM_INVADERS_HORIZONTAL; col++ ) {
			if( p_invaders[row][col].state == ALIVE ) {
				result++;
			}
		}
	}
	
	return result;
}

void p_cursorRelocate() {
	int row, col;
	int aliveInvaderCount = invaders_countRemaining();
	int remainingUntilChosen = rand() % aliveInvaderCount;
	for( row = 0; row < NUM_INVADERS_VERTICAL; row++ ) {
		for( col = 0; col < NUM_INVADERS_HORIZONTAL; col++ ) {
		
			if( p_invaders[row][col].state != ALIVE ) {
				continue;
			}
		
			if( remainingUntilChosen == 0 ) {
				// relocate
				p_cursorRow = row;
				p_cursorCol = col;
				p_cursorFramesSinceMoved = 0;
				return;
			}
			
			remainingUntilChosen--;
		}
	}
}

int32 p_getInvaderXFromCol( int col ) {
	int32 x;
	x = col * ( INVADER_SPRITE_SIZE + BUFFER_BETWEEN_INVADER_SPRITES );
	x += MOVEMENT_SCREEN_BOUNDARY_HORIZONTAL_PIXELS;
	return x;
}

int32 p_getInvaderYFromRow( int row ) {
	int32 y;
	y = row * ( INVADER_SPRITE_SIZE + BUFFER_BETWEEN_INVADER_SPRITES );
	y += MOVEMENT_SCREEN_BOUNDARY_VERTICAL_PIXELS;
	return y;
}

void p_cursorRefreshPosition() {
	
	int x, y;
	int offset = (CURSOR_SPRITE_SIZE - INVADER_SPRITE_SIZE) / 2;
	
	p_cursorFramesSinceMoved++;
	if( p_cursorFramesSinceMoved >= CURSOR_FRAME_COUNT_TO_STAY_PUT ) {
		p_cursorRelocate();
	}
	
	x = p_invaders[p_cursorRow][p_cursorCol].x - offset;
	y = p_invaders[p_cursorRow][p_cursorCol].y - offset;
	
	sprite_moveTo( p_cursor, x, y );
}

void p_cursorInit() {
	p_cursor = sprite_load( p_cursorFilepath );
	display_displayListAdd( 
		(Node*)p_cursor, 
		TRUE, 
		sprite_allocateNextIndex()
		);
}

void invaders_loadDefaults() {
	int row, col;
	Animatable* current;
	
	p_movementCounterVideoFrames = 0;
	p_cursorFramesSinceMoved = 0;
	p_movementIsToTheRight = TRUE;
	
	for( row = 0; row < NUM_INVADERS_VERTICAL; row++ ) {
		for( col = 0; col < NUM_INVADERS_HORIZONTAL; col++ ) {
			current = &p_invaders[row][col];
			
			current->x = p_getInvaderXFromCol( col );
			current->y = p_getInvaderYFromRow( row );			
			
			current->state = ALIVE;
			current->firstFrame = FIRST_MOVEMENT_FRAME;
			current->lastFrame = LAST_MOVEMENT_FRAME;
			current->currentFrame = (row+col) % (LAST_MOVEMENT_FRAME + 1 );
			current->videoFrame = 0;
			
			sprite_moveTo( current->sprite, current->x, current->y );
			sprite_setFrame( current->sprite, current->currentFrame );
		}
	}
	
	p_cursorRelocate();
	sprite_hide( p_cursor );
}

void p_infoPanelInit() {
	p_infoPanel = sprite_load( p_infoPanelFilepath );
	display_displayListAdd( 
		(Node*)p_infoPanel, 
		TRUE, 
		sprite_allocateNextIndex()
		);
	sprite_moveTo( p_infoPanel, display_getScreenWidth() - 100, display_getScreenHeight() - 48 );
}

void invaders_init(
	bool (*checkCollisionCallback)( Rectangle position ) 
	) {
	int row, col;
	Animatable* current;
	
	p_invadersCheckCollisionCallback = checkCollisionCallback;
	
	for( row = 0; row < NUM_INVADERS_VERTICAL; row++ ) {
		for( col = 0; col < NUM_INVADERS_HORIZONTAL; col++ ) {
			current = &p_invaders[row][col];
			
			current->sprite = sprite_load( p_invaderFilepaths[ (row+col) % NUM_INVADER_TEXTURES ] );
			current->spriteIndex = sprite_allocateNextIndex();
			sprite_setFrame( current->sprite, 0 );

			display_displayListAdd( 
				(Node*)current->sprite, 
				TRUE, 
				current->spriteIndex
				);
		}
	}
	p_infoPanelInit();
	p_cursorInit();
	p_invaderBulletHandle = bullets_allocate();
	
	invaders_loadDefaults();
}

void invaders_forceCursorMove() {
	p_cursorFramesSinceMoved = CURSOR_FRAME_COUNT_TO_STAY_PUT;
	p_cursorRefreshPosition();
}

void invaders_refresh() {
	p_moveInvaders();
	p_animateInvaders();
	
	p_cursorRefreshPosition();
}

bool invaders_checkCollision( Rectangle position ) {
	int row, col;
	Rectangle hitbox;
	
	for( row = 0; row < NUM_INVADERS_VERTICAL; row++ ) {
		for( col = 0; col < NUM_INVADERS_HORIZONTAL; col++ ) {
		
			if( p_invaders[row][col].state != ALIVE ) {
				continue;
			}
		
			hitbox = p_invaders[row][col].sprite->position;
			hitbox.topLeft.pt_X += 3;
			hitbox.bottomRight.pt_X -= 3;
			if( sprite_isRectangleOverlap( &position, &hitbox ) ) {
				p_explodeInvader( &p_invaders[row][col] );
				return TRUE;
			}
		}
	}

	return FALSE;
}

void p_invadersCheckCollisionAndHideBullet( Rectangle position ) {
	if( p_invadersCheckCollisionCallback( position ) ) {
		bullets_cancel( p_invaderBulletHandle );
	}
}

void invaders_fire() {
	
	if( p_invaders[p_cursorRow][p_cursorCol].state != ALIVE ) {
		return;
	}
	
	bullets_fire( 
		p_invaderBulletHandle, 
		SLOW,
		DOWN, 
		p_invaders[p_cursorRow][p_cursorCol].x + INVADER_SPRITE_SIZE / 2 - BULLET_SPRITE_SIZE / 2,
		p_invaders[p_cursorRow][p_cursorCol].y + INVADER_SPRITE_SIZE,
		p_invadersCheckCollisionAndHideBullet
		);
}
